/*
 *
 * Official Python API
 *
 * Copyright (c) 2009-2021, quasardb SAS. All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of quasardb nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#pragma once

#include "handle.hpp"
#include <qdb/direct.h>
#include <memory>
#include <string>

namespace qdb
{

class direct_handle
{
public:
    explicit direct_handle() noexcept
    {}

    ~direct_handle()
    {
        if (_handle)
        {
            qdb_direct_close(_handle);
            _handle = nullptr;
        }
    }

    void connect(const handle_ptr handle, const std::string & uri)
    {
        _handle = qdb_direct_connect(*handle, uri.c_str());
        if (_handle == nullptr)
        {
            // Throw here to detect an error as soon as possible.
            // Otherwise, C API would throw only when using the null handle.
            qdb_throw_if_error(*handle, qdb_e_invalid_argument);
        }
    }

    operator qdb_direct_handle_t() const noexcept
    {
        return _handle;
    }

private:
    qdb_direct_handle_t _handle{nullptr};
};

using direct_handle_ptr = std::shared_ptr<direct_handle>;

static inline direct_handle_ptr make_direct_handle_ptr()
{
    return std::make_shared<qdb::direct_handle>(direct_handle());
}

} // namespace qdb
